@@ -1,6 +1,6 @@
 /*
   This file is part of Challenger
-  Copyright (C) 2023 Taler Systems SA
+  Copyright (C) 2023, 2024 Taler Systems SA
 
   Challenger is free software; you can redistribute it and/or modify it under the
   terms of the GNU Affero General Public License as published by the Free Software
@@ -26,6 +26,38 @@
 #include "challenger-httpd_common.h"
 
 
+/**
+ * Generate error reply in the format requested by
+ * the client.
+ *
+ * @param hc our context
+ * @param template error template to use
+ * @param http_status HTTP status to return
+ * @param ec error code to return
+ * @param hint human-readable hint to give
+ */
+static MHD_RESULT
+reply_error (struct CH_HandlerContext *hc,
+             const char *template,
+             unsigned int http_status,
+             enum TALER_ErrorCode ec,
+             const char *hint)
+{
+  if (0 == CH_get_output_type (hc->connection))
+    return TALER_TEMPLATING_reply_error (
+      hc->connection,
+      template,
+      http_status,
+      ec,
+      hint);
+  return TALER_MHD_reply_with_error (
+    hc->connection,
+    http_status,
+    ec,
+    hint);
+}
+
+
 MHD_RESULT
 CH_handler_authorize (struct CH_HandlerContext *hc,
                       const char *upload_data,
@@ -47,10 +79,12 @@ CH_handler_authorize (struct CH_HandlerContext *hc,
                                      sizeof (nonce)))
   {
     GNUNET_break_op (0);
-    return TALER_MHD_reply_with_error (hc->connection,
-                                       MHD_HTTP_NOT_FOUND,
-                                       TALER_EC_GENERIC_PARAMETER_MISSING,
-                                       hc->path);
+    return reply_error (
+      hc,
+      "invalid-request",
+      MHD_HTTP_NOT_FOUND,
+      TALER_EC_GENERIC_PARAMETER_MISSING,
+      hc->path);
   }
   response_type
     = MHD_lookup_connection_value (hc->connection,
@@ -59,19 +93,23 @@ CH_handler_authorize (struct CH_HandlerContext *hc,
   if (NULL == response_type)
   {
     GNUNET_break_op (0);
-    return TALER_MHD_reply_with_error (hc->connection,
-                                       MHD_HTTP_BAD_REQUEST,
-                                       TALER_EC_GENERIC_PARAMETER_MISSING,
-                                       "response_type");
+    return reply_error (
+      hc,
+      "invalid-request",
+      MHD_HTTP_BAD_REQUEST,
+      TALER_EC_GENERIC_PARAMETER_MISSING,
+      "response_type");
   }
   if (0 != strcmp (response_type,
                    "code"))
   {
     GNUNET_break_op (0);
-    return TALER_MHD_reply_with_error (hc->connection,
-                                       MHD_HTTP_BAD_REQUEST,
-                                       TALER_EC_GENERIC_PARAMETER_MALFORMED,
-                                       "response_type (must be 'code')");
+    return reply_error (
+      hc,
+      "invalid-request",
+      MHD_HTTP_BAD_REQUEST,
+      TALER_EC_GENERIC_PARAMETER_MALFORMED,
+      "response_type (must be 'code')");
   }
 
   {
@@ -85,10 +123,12 @@ CH_handler_authorize (struct CH_HandlerContext *hc,
     if (NULL == client_id_str)
     {
       GNUNET_break_op (0);
-      return TALER_MHD_reply_with_error (hc->connection,
-                                         MHD_HTTP_BAD_REQUEST,
-                                         TALER_EC_GENERIC_PARAMETER_MISSING,
-                                         "client_id");
+      return reply_error (
+        hc,
+        "invalid_request",
+        MHD_HTTP_BAD_REQUEST,
+        TALER_EC_GENERIC_PARAMETER_MISSING,
+        "client_id");
     }
     if (1 != sscanf (client_id_str,
                      "%llu%c",
@@ -96,10 +136,12 @@ CH_handler_authorize (struct CH_HandlerContext *hc,
                      &dummy))
     {
       GNUNET_break_op (0);
-      return TALER_MHD_reply_with_error (hc->connection,
-                                         MHD_HTTP_BAD_REQUEST,
-                                         TALER_EC_GENERIC_PARAMETER_MALFORMED,
-                                         "client_id");
+      return reply_error (
+        hc,
+        "invalid-request",
+        MHD_HTTP_BAD_REQUEST,
+        TALER_EC_GENERIC_PARAMETER_MALFORMED,
+        "client_id");
     }
   }
   redirect_uri
@@ -120,10 +162,12 @@ CH_handler_authorize (struct CH_HandlerContext *hc,
                       strlen ("https://"))) )
   {
     GNUNET_break_op (0);
-    return TALER_MHD_reply_with_error (hc->connection,
-                                       MHD_HTTP_BAD_REQUEST,
-                                       TALER_EC_GENERIC_PARAMETER_MALFORMED,
-                                       "redirect_uri (has to start with 'http://' or 'https://')");
+    return reply_error (
+      hc,
+      "invalid-request",
+      MHD_HTTP_BAD_REQUEST,
+      TALER_EC_GENERIC_PARAMETER_MALFORMED,
+      "redirect_uri (has to start with 'http://' or 'https://')");
   }
   state
     = MHD_lookup_connection_value (hc->connection,
@@ -131,6 +175,7 @@ CH_handler_authorize (struct CH_HandlerContext *hc,
                                    "state");
   if (NULL == state)
     state = "";
+
   scope
     = MHD_lookup_connection_value (hc->connection,
                                    MHD_GET_ARGUMENT_KIND,
@@ -158,11 +203,12 @@ CH_handler_authorize (struct CH_HandlerContext *hc,
     {
     case GNUNET_DB_STATUS_HARD_ERROR:
       GNUNET_break (0);
-      return TALER_TEMPLATING_reply_error (hc->connection,
-                                           "internal-error",
-                                           MHD_HTTP_INTERNAL_SERVER_ERROR,
-                                           TALER_EC_GENERIC_DB_STORE_FAILED,
-                                           "authorize_start");
+      return reply_error (
+        hc,
+        "internal-error",
+        MHD_HTTP_INTERNAL_SERVER_ERROR,
+        TALER_EC_GENERIC_DB_STORE_FAILED,
+        "authorize_start");
     case GNUNET_DB_STATUS_SOFT_ERROR:
       GNUNET_break (0);
       return MHD_NO;
@@ -172,11 +218,12 @@ CH_handler_authorize (struct CH_HandlerContext *hc,
                   "Failed to find authorization process of client %llu for nonce `%s'\n",
                   client_id,
                   hc->path);
-      return TALER_TEMPLATING_reply_error (hc->connection,
-                                           "validation-unknown",
-                                           MHD_HTTP_NOT_FOUND,
-                                           TALER_EC_CHALLENGER_GENERIC_VALIDATION_UNKNOWN,
-                                           NULL);
+      return reply_error (
+        hc,
+        "validation-unknown",
+        MHD_HTTP_NOT_FOUND,
+        TALER_EC_CHALLENGER_GENERIC_VALIDATION_UNKNOWN,
+        NULL);
     case GNUNET_DB_STATUS_SUCCESS_ONE_RESULT:
       break;
     }
@@ -221,22 +268,32 @@ CH_handler_authorize (struct CH_HandlerContext *hc,
                                  address_attempts_left)
         );
       http_status = MHD_HTTP_OK;
-      ret = TALER_TEMPLATING_build (
-        hc->connection,
-        &http_status,
-        form,
-        NULL,
-        NULL,
-        args,
-        &resp);
-      GNUNET_free (form);
-      json_decref (args);
-      if (GNUNET_SYSERR == ret)
+      if (0 == CH_get_output_type (hc->connection))
       {
-        GNUNET_break (0);
-        return MHD_NO;
+        ret = TALER_TEMPLATING_build (
+          hc->connection,
+          &http_status,
+          form,
+          NULL,
+          NULL,
+          args,
+          &resp);
+        GNUNET_free (form);
+        json_decref (args);
+        if (GNUNET_SYSERR == ret)
+        {
+          GNUNET_break (0);
+          return MHD_NO;
+        }
+        GNUNET_break (GNUNET_OK == ret);
+      }
+      else
+      {
+        GNUNET_break (0 ==
+                      json_object_del (args,
+                                       "nonce"));
+        resp = TALER_MHD_make_json_steal (args);
       }
-      GNUNET_break (GNUNET_OK == ret);
       GNUNET_break (MHD_YES ==
                     MHD_add_response_header (resp,
                                              MHD_HTTP_HEADER_CACHE_CONTROL,
