@@ -1,6 +1,6 @@
 /*
   This file is part of Challenger
-  Copyright (C) 2023 Taler Systems SA
+  Copyright (C) 2023, 2024 Taler Systems SA
 
   Challenger is free software; you can redistribute it and/or modify it under the
   terms of the GNU Affero General Public License as published by the Free Software
@@ -20,12 +20,14 @@
  */
 #include "platform.h"
 #include "challenger-httpd.h"
+#include <regex.h>
 #include <gnunet/gnunet_util_lib.h>
 #include "challenger-httpd_challenge.h"
 #include <taler/taler_json_lib.h>
 #include <taler/taler_templating_lib.h>
 #include <taler/taler_merchant_service.h>
 #include <taler/taler_signatures.h>
+#include "challenger-httpd_common.h"
 
 
 /**
@@ -84,6 +86,16 @@ struct ChallengeContext
    */
   char *data;
 
+  /**
+   * Where to redirect the client on errors?
+   */
+  char *client_redirect_uri;
+
+  /**
+   * OAuth2 state.
+   */
+  char *state;
+
   /**
    * When did we transmit last?
    */
@@ -129,6 +141,12 @@ struct ChallengeContext
    */
   enum GNUNET_GenericReturnValue suspended;
 
+  /**
+   * True if the provided address was refused, usually because
+   * the user tried too many different addresses already.
+   */
+  bool address_refused;
+
   /**
    * Should we retransmit the PIN?
    */
@@ -198,11 +216,47 @@ cleanup_ctx (void *cls)
   }
   json_decref (bc->address);
   GNUNET_free (bc->data);
+  GNUNET_free (bc->state);
   GNUNET_free (bc->last_key);
+  GNUNET_free (bc->client_redirect_uri);
   GNUNET_free (bc);
 }
 
 
+/**
+ * Generate error reply in the format requested by
+ * the client.
+ *
+ * @param bc our context
+ * @param template error template to use
+ * @param http_status HTTP status to return
+ * @param ec error code to return
+ * @param hint human-readable hint to give
+ */
+static MHD_RESULT
+reply_error (struct ChallengeContext *bc,
+             const char *template,
+             unsigned int http_status,
+             enum TALER_ErrorCode ec,
+             const char *hint)
+{
+  struct CH_HandlerContext *hc = bc->hc;
+
+  if (0 == CH_get_output_type (hc->connection))
+    return TALER_TEMPLATING_reply_error (
+      hc->connection,
+      template,
+      http_status,
+      ec,
+      hint);
+  return TALER_MHD_reply_with_error (
+    hc->connection,
+    http_status,
+    ec,
+    hint);
+}
+
+
 /**
  * Function called when our PIN transmission helper has terminated.
  *
@@ -250,11 +304,11 @@ send_tan (struct ChallengeContext *bc)
     MHD_RESULT mres;
 
     GNUNET_break (0);
-    mres = TALER_TEMPLATING_reply_error (bc->hc->connection,
-                                         "internal-error",
-                                         MHD_HTTP_INTERNAL_SERVER_ERROR,
-                                         TALER_EC_CHALLENGER_HELPER_EXEC_FAILED,
-                                         "pipe");
+    mres = reply_error (bc,
+                        "internal-error",
+                        MHD_HTTP_INTERNAL_SERVER_ERROR,
+                        TALER_EC_CHALLENGER_HELPER_EXEC_FAILED,
+                        "pipe");
     bc->status = (MHD_YES == mres)
       ? GNUNET_NO
       : GNUNET_SYSERR;
@@ -285,11 +339,11 @@ send_tan (struct ChallengeContext *bc)
 
     GNUNET_break (0);
     GNUNET_DISK_pipe_close (p);
-    mres = TALER_TEMPLATING_reply_error (bc->hc->connection,
-                                         "internal-error",
-                                         MHD_HTTP_INTERNAL_SERVER_ERROR,
-                                         TALER_EC_CHALLENGER_HELPER_EXEC_FAILED,
-                                         "exec");
+    mres = reply_error (bc,
+                        "internal-error",
+                        MHD_HTTP_INTERNAL_SERVER_ERROR,
+                        TALER_EC_CHALLENGER_HELPER_EXEC_FAILED,
+                        "exec");
     bc->status = (MHD_YES == mres)
       ? GNUNET_NO
       : GNUNET_SYSERR;
@@ -319,11 +373,11 @@ send_tan (struct ChallengeContext *bc)
         MHD_RESULT mres;
 
         GNUNET_break (0);
-        mres = TALER_TEMPLATING_reply_error (bc->hc->connection,
-                                             "internal-error",
-                                             MHD_HTTP_INTERNAL_SERVER_ERROR,
-                                             TALER_EC_CHALLENGER_HELPER_EXEC_FAILED,
-                                             "write");
+        mres = reply_error (bc,
+                            "internal-error",
+                            MHD_HTTP_INTERNAL_SERVER_ERROR,
+                            TALER_EC_CHALLENGER_HELPER_EXEC_FAILED,
+                            "write");
         GNUNET_free (msg);
         bc->status = (MHD_YES == mres)
           ? GNUNET_NO
@@ -410,6 +464,61 @@ post_iter (void *cls,
 }
 
 
+/**
+ * Check if the given address satisfies our restrictions.
+ *
+ * @param address address data provided by the client
+ * @return NULL on success, otherwise the key that failed
+ */
+static const char *
+check_restrictions (const json_t *address)
+{
+  const char *key;
+  const json_t *val;
+
+  json_object_foreach ((json_t *) address, key, val)
+  {
+    const char *str = json_string_value (val);
+    const char *regex = json_string_value (
+      json_object_get (
+        json_object_get (CH_restrictions,
+                         key),
+        "regex"));
+    regex_t re;
+
+    if (NULL == str)
+      return key;
+    if (NULL == regex)
+      continue;
+    if (0 != regcomp (&re,
+                      regex,
+                      REG_EXTENDED))
+    {
+      GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
+                  "Invalid regex `%s' address restriction specified for `%s'\n",
+                  regex,
+                  key);
+      continue;
+    }
+    if (0 != regexec (&re,
+                      str,
+                      0,
+                      NULL,
+                      0))
+    {
+      GNUNET_log (GNUNET_ERROR_TYPE_WARNING,
+                  "Client input `%s' rejected as it does not match address restriction `%s' specified for `%s'\n",
+                  str,
+                  regex,
+                  key);
+      return key;
+    }
+    regfree (&re);
+  }
+  return NULL;
+}
+
+
 MHD_RESULT
 CH_handler_challenge (struct CH_HandlerContext *hc,
                       const char *upload_data,
@@ -441,11 +550,11 @@ CH_handler_challenge (struct CH_HandlerContext *hc,
                                        sizeof (bc->nonce)))
     {
       GNUNET_break_op (0);
-      return TALER_TEMPLATING_reply_error (hc->connection,
-                                           "invalid-request",
-                                           MHD_HTTP_NOT_FOUND,
-                                           TALER_EC_GENERIC_PARAMETER_MISSING,
-                                           hc->path);
+      return reply_error (bc,
+                          "invalid-request",
+                          MHD_HTTP_NOT_FOUND,
+                          TALER_EC_GENERIC_PARAMETER_MISSING,
+                          hc->path);
     }
     TALER_MHD_check_content_length (hc->connection,
                                     1024);
@@ -473,11 +582,11 @@ CH_handler_challenge (struct CH_HandlerContext *hc,
                      "%u/%d",
                      (unsigned int) bc->exit_code,
                      bc->pst);
-    return TALER_TEMPLATING_reply_error (hc->connection,
-                                         "internal-error",
-                                         MHD_HTTP_INTERNAL_SERVER_ERROR,
-                                         TALER_EC_CHALLENGER_HELPER_EXEC_FAILED,
-                                         es);
+    return reply_error (bc,
+                        "internal-error",
+                        MHD_HTTP_INTERNAL_SERVER_ERROR,
+                        TALER_EC_CHALLENGER_HELPER_EXEC_FAILED,
+                        es);
   }
   /* handle upload */
   if (0 != *upload_data_size)
@@ -514,49 +623,83 @@ CH_handler_challenge (struct CH_HandlerContext *hc,
                 address);
     free (address);
   }
+  {
+    const char *bad_field;
+
+    bad_field = check_restrictions (bc->address);
+    if (NULL != bad_field)
+    {
+      GNUNET_break_op (0);
+      return reply_error (bc,
+                          "invalid-request",
+                          MHD_HTTP_BAD_REQUEST,
+                          TALER_EC_GENERIC_PARAMETER_MALFORMED,
+                          bad_field);
+    }
+  }
   if (! bc->db_finished)
   {
     enum GNUNET_DB_QueryStatus qs;
 
-    qs = CH_db->challenge_set_address_and_pin (CH_db->cls,
-                                               &bc->nonce,
-                                               bc->address,
-                                               CH_validation_duration,
-                                               &bc->tan,
-                                               &bc->last_tx_time,
-                                               &bc->pin_attempts_left,
-                                               &bc->retransmit);
+    GNUNET_assert (NULL == bc->client_redirect_uri);
+    qs = CH_db->challenge_set_address_and_pin (
+      CH_db->cls,
+      &bc->nonce,
+      bc->address,
+      CH_validation_duration,
+      &bc->tan,
+      &bc->state,
+      &bc->last_tx_time,
+      &bc->pin_attempts_left,
+      &bc->retransmit,
+      &bc->client_redirect_uri,
+      &bc->address_refused);
     switch (qs)
     {
     case GNUNET_DB_STATUS_HARD_ERROR:
       GNUNET_break (0);
-      return TALER_TEMPLATING_reply_error (hc->connection,
-                                           "internal-error",
-                                           MHD_HTTP_NOT_FOUND,
-                                           TALER_EC_GENERIC_DB_STORE_FAILED,
-                                           "set-address-and-pin");
+      return reply_error (bc,
+                          "internal-error",
+                          MHD_HTTP_INTERNAL_SERVER_ERROR,
+                          TALER_EC_GENERIC_DB_STORE_FAILED,
+                          "set-address-and-pin");
     case GNUNET_DB_STATUS_SOFT_ERROR:
       GNUNET_break (0);
       return MHD_NO;
     case GNUNET_DB_STATUS_SUCCESS_NO_RESULTS:
-      return TALER_TEMPLATING_reply_error (hc->connection,
-                                           "validation-unknown",
-                                           MHD_HTTP_NOT_FOUND,
-                                           TALER_EC_CHALLENGER_GENERIC_VALIDATION_UNKNOWN,
-                                           NULL);
+      GNUNET_break_op (0);
+      return reply_error (bc,
+                          "validation-unknown",
+                          MHD_HTTP_NOT_FOUND,
+                          TALER_EC_CHALLENGER_GENERIC_VALIDATION_UNKNOWN,
+                          NULL);
     case GNUNET_DB_STATUS_SUCCESS_ONE_RESULT:
       break;
     }
     bc->db_finished = true;
+    if (bc->address_refused)
+    {
+      GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                  "Address changes exhausted address change limit for this process\n");
+      return TALER_MHD_redirect_with_oauth_status (
+        hc->connection,
+        bc->client_redirect_uri,
+        bc->state,
+        "unauthorized_client",
+        "client exceeded authorization attempts limit (too many addresses attempted)",
+        NULL);
+    }
     if (0 == bc->pin_attempts_left)
     {
       GNUNET_log (GNUNET_ERROR_TYPE_INFO,
-                  "Attempts exhausted for this PIN\n");
-      return TALER_TEMPLATING_reply_error (hc->connection,
-                                           "attempts-exhausted",
-                                           MHD_HTTP_TOO_MANY_REQUESTS,
-                                           TALER_EC_CHALLENGER_TOO_MANY_ATTEMPTS,
-                                           NULL);
+                  "Address changes exhausted PIN limit for this address\n");
+      return TALER_MHD_redirect_with_oauth_status (
+        hc->connection,
+        bc->client_redirect_uri,
+        bc->state,
+        "unauthorized_client",
+        "client exceeded authorization attempts limit (too many PINs)",
+        NULL);
     }
 
     if (bc->retransmit)
@@ -578,6 +721,9 @@ CH_handler_challenge (struct CH_HandlerContext *hc,
   {
     json_t *args;
     enum GNUNET_GenericReturnValue ret;
+    struct MHD_Response *resp;
+    unsigned int http_status;
+    MHD_RESULT res;
 
     args = GNUNET_JSON_PACK (
       GNUNET_JSON_pack_uint64 ("attempts_left",
@@ -594,19 +740,39 @@ CH_handler_challenge (struct CH_HandlerContext *hc,
                                    bc->last_tx_time,
                                    CH_validation_duration)))
       );
-    ret = TALER_TEMPLATING_reply (hc->connection,
-                                  MHD_HTTP_OK,
-                                  "enter-tan-form",
-                                  NULL,
-                                  NULL,
-                                  args);
-    json_decref (args);
-    if (GNUNET_SYSERR == ret)
+    http_status = MHD_HTTP_OK;
+    if (0 == CH_get_output_type (hc->connection))
     {
-      GNUNET_break (0);
-      return MHD_NO;
+      ret = TALER_TEMPLATING_build (hc->connection,
+                                    &http_status,
+                                    "enter-tan-form",
+                                    NULL,
+                                    NULL,
+                                    args,
+                                    &resp);
+      json_decref (args);
+      if (GNUNET_SYSERR == ret)
+      {
+        GNUNET_break (0);
+        return MHD_NO;
+      }
+      GNUNET_break (GNUNET_OK == ret);
     }
-    GNUNET_break (GNUNET_OK == ret);
-    return MHD_YES;
+    else
+    {
+      GNUNET_break (0 ==
+                    json_object_del (args,
+                                     "nonce"));
+      resp = TALER_MHD_make_json (args);
+    }
+    GNUNET_break (MHD_YES ==
+                  MHD_add_response_header (resp,
+                                           MHD_HTTP_HEADER_CACHE_CONTROL,
+                                           "no-store,no-cache"));
+    res = MHD_queue_response (hc->connection,
+                              http_status,
+                              resp);
+    MHD_destroy_response (resp);
+    return res;
   }
 }
