@@ -24,6 +24,25 @@
 
 #define RFC_8959_PREFIX "secret-token:"
 
+
+int
+CH_get_output_type (struct MHD_Connection *connection)
+{
+  int output_type;
+
+  output_type
+    = TALER_MHD_check_accept (connection,
+                              MHD_HTTP_HEADER_ACCEPT,
+                              "text/html",
+                              "application/json",
+                              NULL);
+  if (output_type < 0)
+    output_type = 0; /* default to HTML */
+  GNUNET_assert (output_type < 2);
+  return output_type;
+}
+
+
 const char *
 CH_get_client_secret (struct MHD_Connection *connection)
 {
